# -*- coding: utf-8 -*-
"""
Interaction of classes in `bayesvalidrox`
@author: Rebecca Kohlhaas
"""
import graphviz
from pyUML import Graph, UMLClass

def print_and_save(graph, name):
    #print(graph.to_string())
    #graph.write_raw(f"{name}.dot")
    graph.write(f"{name}.png")#, "dot", "png")

def generate_input_uml():
    # Create UML for the input classes
    graph = Graph('pyUML')
    
    # Add input class
    inputs = UMLClass('Input',
                      attributes={
                              'Marginals':'list',
                              'Rosenblatt':'bool'
                          },
                      methods=['add_marginals()']
                      )
    graph.add_class(inputs)
    
    # Add marginal class
    marginals = UMLClass('Marginal',
                         attributes={
                                 'name':'string',
                                 'dist_type': 'string',
                                 'parameters': 'list',
                                 'input_data':'array',
                                 'moments': 'list'
                             }
                         )
    graph.add_class(marginals)
    graph.add_composition(marginals, inputs, multiplicity_parent=1, multiplicity_child='1..*')
    
    inputspace = UMLClass('InputSpace',
                          attributes = {
                              'input_obj':'Input',
                              'meta_Model_type':'string'
                              },
                          methods = [
                              'check_valid_inputs()', 'init_param_space()', 
                              'build_polytypes()', 'transform()'
                              ]
                          )
    graph.add_class(inputspace)
    graph.add_composition(inputs, inputspace, multiplicity_parent=1, multiplicity_child=1)
    
    expdesign = UMLClass('ExpDesigns',
                         attributes = {
                             'sampling_method': 'string',
                             'hdf5_file': 'string',
                             'n_new_samples': 'int',
                             'n_max_samples': 'int',
                             'mod_LOO_threshold': 'double',
                             'tradeoff_scheme':'string',
                             'n_canddidate': 'int',
                             'explore_method':'string',
                             'exploit_method': 'string',
                             'util_func':'string',
                             'n_cand_groups': 'int',
                             'n_replication': 'int',
                             'post_snapshot': 'bool', 
                             'step_snapshot':'int',
                             'max_a_post': 'list',
                             'adapt_verbose':'bool',
                             'max_func_itr': 'int',
                             'out_dir': 'string'
                             },
                         methods = [
                             'generate_samples()', 'generate_ED()', 
                             'read_from_file()', 'random_sampler()',
                             'pcm_sampler()', 'plot_samples()'
                             ])
    graph.add_class(expdesign)
    graph.add_implementation(expdesign, inputspace)
    
    print_and_save(graph, 'input_classes')
    
def generate_input_uml_reduced():
    # Create UML for the input classes
    graph = Graph('pyUML')
    
    # Add input class
    inputs = UMLClass('Input',
                      attributes={
                              'Marginals':'list',
                              'Rosenblatt':'bool'
                          },
                      methods=['add_marginals()']
                      )
    graph.add_class(inputs)
    
    # Add marginal class
    marginals = UMLClass('Marginal',
                         attributes={
                                 'name':'string',
                                 'dist_type': 'string',
                                 'parameters': 'list',
                                 'input_data':'array',
                                 'moments': 'list'
                             }
                         )
    graph.add_class(marginals)
    graph.add_composition(marginals, inputs, multiplicity_parent=1, multiplicity_child='1..*')
    
    inputspace = UMLClass('InputSpace',
                          attributes = {
                              'input_obj':'Input',
                              'meta_Model_type':'string'
                              },
                          methods = [
                              'check_valid_inputs()', 'init_param_space()', 
                              'transform()'
                              ]
                          )
    graph.add_class(inputspace)
    graph.add_composition(inputs, inputspace, multiplicity_parent=1, multiplicity_child=1)
    
    expdesign = UMLClass('ExpDesigns',
                         attributes = {
                             'sampling_method': 'string',
                             'hdf5_file': 'string',
                             'out_dir': 'string'
                             },
                         methods = [
                             'generate_samples()', 'generate_ED()', 
                             'read_from_file()', 'random_sampler()',
                             'pcm_sampler()', 'plot_samples()'
                             ])
    graph.add_class(expdesign)
    graph.add_implementation(expdesign, inputspace)
    
    print_and_save(graph, 'input_classes_reduced')
    
def generate_training_uml():
    """
    Generates the uml for static training and related classes
    """
    graph = Graph('pyUML')
    
    # Add input class
    inputs = UMLClass('Input')
    graph.add_class(inputs)
    inputspace = UMLClass('InputSpace')
    graph.add_class(inputspace)
    
    expdesign = UMLClass('ExpDesigns',
                         attributes = {
                             'sampling_method': 'string',
                             'hdf5_file': 'string',
                             'n_new_samples': 'int',
                             'n_max_samples': 'int',
                             'mod_LOO_threshold': 'double',
                             'tradeoff_scheme':'string',
                             'n_canddidate': 'int',
                             'explore_method':'string',
                             'exploit_method': 'string',
                             'util_func':'string',
                             'n_cand_groups': 'int',
                             'n_replication': 'int',
                             'post_snapshot': 'bool', 
                             'step_snapshot':'int',
                             'max_a_post': 'list',
                             'adapt_verbose':'bool',
                             'max_func_itr': 'int',
                             'out_dir': 'string'
                             },
                         methods = [
                             'generate_samples()', 'generate_ED()', 
                             'read_from_file()', 'random_sampler()',
                             'pcm_sampler()', 'plot_samples()'
                             ])
    graph.add_class(expdesign)
    graph.add_composition(inputs, expdesign, multiplicity_parent = 1, multiplicity_child = 1)
    
    model = UMLClass('PyLinkForwardModel',
                     attributes = {
                         'link_type':'string', 
                         'name':'string', 
                         'py_file':'string',
                        'func_args':'dict', 
                        'shell_command':'string', 
                        'input_file':'string',
                        'input_template':'string', 
                        'aux_file':'string', 
                        'exe_path':'string',
                        'output_file_names':'list', 
                        'output_names':'list', 
                        'output_parser':'string',
                        'multi_process':'bool', 
                        'n_cpus':'int', 
                        'meas_file':'string',
                        'meas_file_valid':'string', 
                        'mc_ref_file':'string', 
                        'obs_dict':'dict',
                        'obs_dict_valid':'dict', 
                        'mc_ref_dict':'dict'
                         },
                     methods = [
                         'read_observation()', 'read_output()','update_input_params()',
                         'run_command()', 'run_forwardmodel()', 'run_model_parallel()',
                         'uMBridge_model()', '_store_simulations()', 'zip_subdirs()'
                         ]
                     )
    graph.add_class(model)
    
    metamod = UMLClass('MetaModel',
                       attributes = {
                           'input_obj':'Input',
                           'meta_model_type': 'string',
                           'pce_reg_method':'string',
                           'bootstrap_method':'string',
                           'n_bootstrap_itrs':'int',
                           'pce_deg':'int',
                           'pce_q_norm':'double',
                           'dim_red_method':'string',
                           'apply_constraints':'bool',
                           'verbose':'bool'
                           },
                       methods = [
                           'build_metamodel()', 'fit()', 'update_pce_coeffs()',
                           'add_InputSpace()', 'univ_basis_vals()', 'regression()',
                           'adaptive_regression()', 'pca_transformation()', 'eval_metamodel()',
                           'create_model_error()', 'eval_model_error()', 'copy_meta_model_opts()', 
                           '_select_degree()', 'generate_polynomials()', '_compute_pce_moments()'
                           ]
                       )
    graph.add_class(metamod)
    graph.add_composition(inputs, metamod, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(inputspace, metamod, multiplicity_parent = 1, multiplicity_child = 1)
    
    engine = UMLClass('Engine',
                      attributes = {
                          'MetaModel':'MetaModel',
                          'Model':'Model',
                          'ExpDesign':'ExpDesigns',
                          'parallel': 'bool', 
                          'trained':'bool',
                          },
                      methods = [
                          'start_engine()', 'train_normal()', 'train_sequential()', 
                          'eval_metamodel()', 'train_seq_design()', 'util_VarBasedDesign()', 
                          'util_BayesianActiveDesign()', 'util_BayesianDesign()', 'run_util_func()', 
                          'dual_annealing()', 'tradeoff_weights()', 'choose_next_sample()', 
                          'util_AlphOptDesign()', '_normpdt()', '_corr_factor_BME()', 
                          '_posteriorPlot()', '_BME_Calculator()', '_validError()', '_error_Mean_Std()'
                          ]
                      )
    graph.add_class(engine)
    graph.add_composition(model, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(expdesign, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(metamod, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    
    # Add the regression Classes
    Breg = UMLClass('BayesianLinearRegression',
                      attributes = {
                          'n_iter':'int',
                          'tol': 'double',
                          'fit_intercept':'bool',
                          'copy_X':'bool',
                          'verbose':'bool'
                          },
                      methods = [
                          '_check_convergence()', '_center_data()', 'predict_dist()'
                          ]
                      )
    graph.add_class(Breg)
    vBreg = UMLClass('VBLinearRegression',
                      attributes = {
                          'optimizer':'string',
                          'alpha':'double',
                          'normalize':'bool',
                          'scores': 'list',
                          'perfect_fit_tol': 'double',
                          },
                      methods = [
                          'fit()', 'predict()'
                          ]
                      )
    graph.add_class(vBreg)
    eBreg = UMLClass('EBLinearRegression',
                      attributes = {
                          'a':'double',
                        'b':'double',
                        'c':'double',
                        'd':'double'
                          },
                      methods = [
                          'fit()', 'predict()', 'postrior_weights()'
                          ]
                      )
    graph.add_class(eBreg)
    omp = UMLClass('OrthogonalMatchingPursuit',
                      attributes = {
                      'fit_intercept':'bool',
                      'normalize':'bool',
                      'copy_X':'bool',
                      'verbose':'bool'
                          },
                      methods = [
                          '_preprocess_data()', 'fit()', 'predict()', 'loo_error()',
                          'blockwise_inverse()'
                          ]
                      )
    graph.add_class(omp)
    regArd = UMLClass('RegressionFastARD',
                      attributes = {
                          'n_iter':'int',
                          'start':'list',
                          'tol':'double',
                          'fit_intercept':'bool',
                          'normalize':'bool',
                          'copy_X':'bool',
                          'compute_score':'bool',
                          'verbose':'bool'
                          },
                      methods = [
                          '_preprocess_data()', 'fit()', 'log_marginal_like()',
                          'predict()', '_posterior_dist()', '_sparsity_quality()'
                          ]
                      )
    graph.add_class(regArd)
    regLap = UMLClass('RegressionFastLaplace',
                      attributes = {
                          'n_iter':'int',
                          'n_Kfold':'int',
                          'tol':'double',
                          'fit_intercept':'bool',
                          'bias_term':'bool',
                          'copy_X':'bool',
                          'verbose':'bool'
                          },
                      methods = [
                          '_center_data()', 'fit()', '_fit()', 'predict()'
                          ]
                      )
    graph.add_class(regLap)
    graph.add_implementation(vBreg, Breg)
    graph.add_implementation(eBreg, Breg)
    graph.add_composition(vBreg, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    graph.add_composition(eBreg, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    graph.add_composition(omp, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    graph.add_composition(regArd, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    graph.add_composition(regLap, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    
    print_and_save(graph, 'metamod_training')
    
def generate_training_uml_reduced():
    """
    Generates the uml for static training and related classes
    """
    graph = Graph('pyUML')
    
    # Add input class
    inputs = UMLClass('Input')
    graph.add_class(inputs)
    inputspace = UMLClass('InputSpace',
                          methods = ['build_polytypes()'])
    graph.add_class(inputspace)
    
    expdesign = UMLClass('ExpDesigns',
                         attributes = {
                             'sampling_method': 'string',
                             'hdf5_file': 'string',
                             'n_init_samples': 'int',
                             'out_dir': 'string'
                             },
                         methods = [
                             'generate_samples()', 'generate_ED()', 
                             'read_from_file()', 'random_sampler()',
                             'pcm_sampler()', 'plot_samples()'
                             ])
    graph.add_class(expdesign)
    graph.add_composition(inputs, expdesign, multiplicity_parent = 1, multiplicity_child = 1)
    
    model = UMLClass('PyLinkForwardModel')
    graph.add_class(model)
    
    metamod = UMLClass('MetaModel',
                       attributes = {
                           'input_obj':'Input',
                           'meta_model_type': 'string',
                           'pce_reg_method':'string',
                           'bootstrap_method':'string',
                           'n_bootstrap_itrs':'int',
                           'pce_deg':'int',
                           'pce_q_norm':'double',
                           'dim_red_method':'string',
                           #'apply_constraints':'bool',
                           'verbose':'bool'
                           },
                       methods = [
                           'build_metamodel()', 'fit()', 'update_pce_coeffs()',
                           'add_InputSpace()', 'univ_basis_vals()', 'regression()',
                           'adaptive_regression()', 'pca_transformation()', 'eval_metamodel()',
                           'copy_meta_model_opts()', '_select_degree()', 
                           'generate_polynomials()', '_compute_pce_moments()'
                           ]
                       )
    graph.add_class(metamod)
    graph.add_composition(inputs, metamod, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(inputspace, metamod, multiplicity_parent = 1, multiplicity_child = 1)
    
    engine = UMLClass('Engine',
                      attributes = {
                          'MetaModel':'MetaModel',
                          'Model':'Model',
                          'ExpDesign':'ExpDesigns',
                          'parallel': 'bool', 
                          'trained':'bool',
                          },
                      methods = [
                          'start_engine()', 'train_normal()',
                          'eval_metamodel()', '_posteriorPlot()', '_validError()',
                          '_error_Mean_Std()'
                          ]
                      )
    graph.add_class(engine)
    graph.add_composition(model, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(expdesign, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(metamod, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    
    # Add the regression Classes
    Breg = UMLClass('BayesianLinearRegression',
                      attributes = {
                          'n_iter':'int',
                          'tol': 'double',
                          'fit_intercept':'bool'
                          },
                      methods = [
                          '_check_convergence()', '_center_data()', 'predict_dist()'
                          ]
                      )
    graph.add_class(Breg)
    vBreg = UMLClass('VBLinearRegression',
                      methods = ['fit()', 'predict()']
                      )
    graph.add_class(vBreg)
    eBreg = UMLClass('EBLinearRegression',
                      methods = ['fit()', 'predict()']
                      )
    graph.add_class(eBreg)
    omp = UMLClass('OrthogonalMatchingPursuit',
                      attributes = {'fit_intercept':'bool'},
                      methods = ['fit()', 'predict()']
                      )
    graph.add_class(omp)
    regArd = UMLClass('RegressionFastARD',
                      attributes = {
                          'n_iter':'int',
                          'tol':'double',
                          'fit_intercept':'bool',
                          'normalize':'bool',
                          'compute_score':'bool'
                          },
                      methods = ['fit()', 'predict()']
                      )
    graph.add_class(regArd)
    regLap = UMLClass('RegressionFastLaplace',
                      attributes = {
                          'n_iter':'int',
                          'tol':'double',
                          'fit_intercept':'bool',
                          'bias_term':'bool'
                          },
                      methods = ['fit()', 'predict()']
                      )
    graph.add_class(regLap)
    graph.add_implementation(vBreg, Breg)
    graph.add_implementation(eBreg, Breg)
    graph.add_composition(vBreg, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    graph.add_composition(eBreg, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    graph.add_composition(omp, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    graph.add_composition(regArd, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    graph.add_composition(regLap, metamod, multiplicity_parent = 1, multiplicity_child = '1..*')
    
    print_and_save(graph, 'metamod_training_reduced')
    
def generate_model_uml():
    
    graph = Graph('pyUML')
    model = UMLClass('PyLinkForwardModel',
                     attributes = {
                         'link_type':'string', 
                         'name':'string', 
                         'py_file':'string',
                        'func_args':'dict', 
                        'shell_command':'string', 
                        'input_file':'string',
                        'input_template':'string', 
                        'aux_file':'string', 
                        'exe_path':'string',
                        'output_file_names':'list', 
                        'output_names':'list', 
                        'output_parser':'string',
                        'multi_process':'bool', 
                        'n_cpus':'int', 
                        'meas_file':'string',
                        'meas_file_valid':'string', 
                        'mc_ref_file':'string', 
                        'obs_dict':'dict',
                        'obs_dict_valid':'dict', 
                        'mc_ref_dict':'dict'
                         },
                     methods = [
                         'read_observation()', 'read_output()','update_input_params()',
                         'run_command()', 'run_forwardmodel()', 'run_model_parallel()',
                         'uMBridge_model()', '_store_simulations()', 'zip_subdirs()'
                         ]
                     )
    graph.add_class(model)
    print_and_save(graph, 'model')
    
def generate_al_uml():
    """
    Generates the uml for active learning
    """
    graph = Graph('pyUML')
    
    # Add input class
    inputs = UMLClass('Input')
    graph.add_class(inputs)
    inputspace = UMLClass('InputSpace')
    graph.add_class(inputspace)
    
    expdesign = UMLClass('ExpDesigns',
                         attributes = {
                             'sampling_method': 'string',
                             'hdf5_file': 'string',
                             'n_new_samples': 'int',
                             'n_max_samples': 'int',
                             'mod_LOO_threshold': 'double',
                             'tradeoff_scheme':'string',
                             'n_canddidate': 'int',
                             'explore_method':'string',
                             'exploit_method': 'string',
                             'util_func':'string',
                             'n_cand_groups': 'int',
                             'n_replication': 'int',
                             'post_snapshot': 'bool', 
                             'step_snapshot':'int',
                             'max_a_post': 'list',
                             'adapt_verbose':'bool',
                             'max_func_itr': 'int',
                             'out_dir': 'string'
                             },
                         methods = [
                             'generate_samples()', 'generate_ED()', 
                             'read_from_file()', 'random_sampler()',
                             'pcm_sampler()', 'plot_samples()'
                             ])
    graph.add_class(expdesign)
    graph.add_composition(inputs, expdesign, multiplicity_parent = 1, multiplicity_child = 1)
    
    model = UMLClass('PyLinkForwardModel',
                     attributes = {
                         'link_type':'string', 
                         'name':'string', 
                         'py_file':'string',
                        'func_args':'dict', 
                        'shell_command':'string', 
                        'input_file':'string',
                        'input_template':'string', 
                        'aux_file':'string', 
                        'exe_path':'string',
                        'output_file_names':'list', 
                        'output_names':'list', 
                        'output_parser':'string',
                        'multi_process':'bool', 
                        'n_cpus':'int', 
                        'meas_file':'string',
                        'meas_file_valid':'string', 
                        'mc_ref_file':'string', 
                        'obs_dict':'dict',
                        'obs_dict_valid':'dict', 
                        'mc_ref_dict':'dict'
                         },
                     methods = [
                         'read_observation()', 'read_output()','update_input_params()',
                         'run_command()', 'run_forwardmodel()', 'run_model_parallel()',
                         'uMBridge_model()', '_store_simulations()', 'zip_subdirs()'
                         ]
                     )
    graph.add_class(model)
    
    metamod = UMLClass('MetaModel',
                       attributes = {
                           'input_obj':'Input',
                           'meta_model_type': 'string',
                           'pce_reg_method':'string',
                           'bootstrap_method':'string',
                           'n_bootstrap_itrs':'int',
                           'pce_deg':'int',
                           'pce_q_norm':'double',
                           'dim_red_method':'string',
                           'apply_constraints':'bool',
                           'verbose':'bool'
                           },
                       methods = [
                           'build_metamodel()', 'fit()', 'update_pce_coeffs()',
                           'add_InputSpace()', 'univ_basis_vals()', 'regression()',
                           'adaptive_regression()', 'pca_transformation()', 'eval_metamodel()',
                           'copy_meta_model_opts()', 
                           '_select_degree()', 'generate_polynomials()', '_compute_pce_moments()'
                           ]
                       )
    graph.add_class(metamod)
    graph.add_composition(inputs, metamod, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(inputspace, metamod, multiplicity_parent = 1, multiplicity_child = 1)
    
    engine = UMLClass('Engine',
                      attributes = {
                          'MetaModel':'MetaModel',
                          'Model':'Model',
                          'ExpDesign':'ExpDesigns',
                          'parallel': 'bool', 
                          'trained':'bool',
                          },
                      methods = [
                          'start_engine()', 'train_normal()', 'train_sequential()', 
                          'eval_metamodel()', 'train_seq_design()', 'util_VarBasedDesign()', 
                          'util_BayesianActiveDesign()', 'util_BayesianDesign()', 'run_util_func()', 
                          'dual_annealing()', 'tradeoff_weights()', 'choose_next_sample()', 
                          'util_AlphOptDesign()', '_normpdt()', '_corr_factor_BME()', 
                          '_posteriorPlot()', '_BME_Calculator()', '_validError()', '_error_Mean_Std()'
                          ]
                      )
    graph.add_class(engine)
    graph.add_composition(model, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(expdesign, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(metamod, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    
    explor = UMLClass('Exploration',
                      attributes = {
                          'ExpDesign':'ExpDesigns',
                          'n_candidate':'int',
                          'mc_criterion':'string',
                          },
                      methods=[
                          'get_exploration_samples()','get_vornoi_samples()',
                          'get_mc_samples()', 'approximate_voronoi()', '_build_dist_matrix_point()'
                          ]
                      )
    graph.add_class(explor)
    graph.add_composition(expdesign, explor, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(explor, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    
    disc = UMLClass('Discrepancy',
                    attributes = {
                        'InputDosc':'Input',
                        'disc_type':'string',
                        'parameters':'list'
                        },
                    methods = [
                        'get_sample()'
                        ]
                    )
    graph.add_class(disc)
    graph.add_composition(expdesign, disc, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(inputs, disc, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(disc, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    print_and_save(graph, 'active_learning')
    
def generate_al_uml_reduced():
    """
    Generates the uml for active learning
    """
    graph = Graph('pyUML')
    
    # Add input class
    inputs = UMLClass('Input')
    graph.add_class(inputs)
    
    expdesign = UMLClass('ExpDesigns',
                         attributes = {
                             'sampling_method': 'string',
                             'hdf5_file': 'string',
                             'n_new_samples': 'int',
                             'n_max_samples': 'int',
                             'mod_LOO_threshold': 'double',
                             'tradeoff_scheme':'string',
                             'n_canddidate': 'int',
                             'explore_method':'string',
                             'exploit_method': 'string',
                             'util_func':'string',
                             'n_cand_groups': 'int',
                             'n_replication': 'int',
                             'post_snapshot': 'bool', 
                             'step_snapshot':'int',
                             'max_a_post': 'list',
                             'adapt_verbose':'bool',
                             'max_func_itr': 'int',
                             'out_dir': 'string'
                             },
                         methods = [
                             'generate_samples()', 'generate_ED()', 
                             'read_from_file()', 'random_sampler()',
                             'pcm_sampler()', 'plot_samples()'
                             ])
    graph.add_class(expdesign)
    
    model = UMLClass('PyLinkForwardModel')
    graph.add_class(model)
    
    metamod = UMLClass('MetaModel')
    graph.add_class(metamod)
    
    engine = UMLClass('Engine',
                      attributes = {
                          'MetaModel':'MetaModel',
                          'Model':'Model',
                          'ExpDesign':'ExpDesigns',
                          'parallel': 'bool', 
                          'trained':'bool',
                          },
                      methods = [
                          'start_engine()', 'train_normal()', 'train_sequential()', 
                          'eval_metamodel()', 'train_seq_design()', '_normpdf()', '_corr_factor_BME()', 
                          '_posteriorPlot()', '_BME_Calculator()', '_validError()', '_error_Mean_Std()'
                          ]
                      )
    graph.add_class(engine)
    graph.add_composition(model, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(expdesign, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(metamod, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    seqdes = UMLClass('SeqDesign',
                      attributes = {
                          'MetaModel':'MetaModel',
                          'Model':'Model',
                          'ExpDesign':'ExpDesigns',
                          'engine': 'Engine'
                          },
                      methods = [
                          'util_VarBasedDesign()', 
                          'util_BayesianActiveDesign()', 'util_BayesianDesign()', 'run_util_func()', 
                          'dual_annealing()', 'tradeoff_weights()', 'choose_next_sample()', 
                          'util_AlphOptDesign()', '_select_indices()', '_error_Mean_Std()', 
                          '_BME_Calculator()', '_validError()', '_normpdf()', '_corr_factor_BME()'
                          ]
                      )
    graph.add_class(seqdes)
    graph.add_composition(seqdes, engine, multiplicity_child=1, multiplicity_parent=1)
    
    explor = UMLClass('Exploration',
                      attributes = {
                          'ExpDesign':'ExpDesigns',
                          'n_candidate':'int',
                          'mc_criterion':'string',
                          },
                      methods=[
                          'get_exploration_samples()','get_vornoi_samples()',
                          'get_mc_samples()', 'approximate_voronoi()', '_build_dist_matrix_point()'
                          ]
                      )
    graph.add_class(explor)
    graph.add_composition(expdesign, explor, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(explor, seqdes, multiplicity_parent = 1, multiplicity_child = 1)
    
    disc = UMLClass('Discrepancy',
                    attributes = {
                        'InputDisc':'Input',
                        'disc_type':'string',
                        'parameters':'list'
                        },
                    methods = [
                        'get_sample()'
                        ]
                    )
    graph.add_class(disc)
    graph.add_composition(expdesign, disc, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(inputs, disc, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(disc, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    print_and_save(graph, 'active_learning_reduced')
    
def generate_postprocessing_uml():
    graph = Graph('pyUML')
    
    expdesign = UMLClass('ExpDesigns')
    graph.add_class(expdesign)
    
    model = UMLClass('PyLinkForwardModel')
    graph.add_class(model)
    
    metamod = UMLClass('MetaModel')
    graph.add_class(metamod)
    
    engine = UMLClass('Engine')
    graph.add_class(engine)
    graph.add_composition(model, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(expdesign, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(metamod, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    post = UMLClass('PostProcessing',
                    attributes = {
                        'engine':'Engine',
                        'name':'string',
                        'out_dir':'string' # TODO: this is somehow not in the code anymore??
                        },
                    methods = [
                        'plot_moments()', 'valid_metamodel()', 'check_accuracy()',
                        'plot_seq_design_diagnostics()', 'sobol_indices()', 
                        'check_reg_quality()', 'eval_pce_model_3d()', 'comput_pce_moments()',
                        '_get_sample()', '_eval_model()', '_plot_validation()', '_plot_validation_multi()'
                        ]
                    )
    graph.add_class(post)
    graph.add_composition(engine, post, multiplicity_parent = 1, multiplicity_child = 1)
    
    print_and_save(graph, 'postprocessing')
    
def generate_bayes_uml():
    graph = Graph('pyUML')
    
    expdesign = UMLClass('ExpDesigns')
    graph.add_class(expdesign)
    
    model = UMLClass('PyLinkForwardModel')
    graph.add_class(model)
    
    metamod = UMLClass('MetaModel')
    graph.add_class(metamod)
    
    engine = UMLClass('Engine')
    graph.add_class(engine)
    graph.add_composition(model, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(expdesign, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(metamod, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    disc = UMLClass('Discrepancy')
    graph.add_class(disc)
    
    bayesinf = UMLClass('BayesInference',
                        attributes = {
                            'engine':'Engine',
                            'discrepancy':'Discrepancy',
                            'emulator':'bool',
                            'name':'string',
                            'bootatrap':'bool',
                            'req_outputs':'list',
                            'selected_indices':'list',
                            'prior_samples':'array',
                            'n_prior_samples':'int',
                            'measured_data':'dict',
                            'inference_method':'string',
                            'mcmc_params':'dict',
                            'bayes_loocv':'bool',
                            'n_bootstrap_itrs':'int',
                            'perturbed_data':'lsit',
                            'bootstrap_noise':'double',
                            'just_analysis':'bool',
                            'valid_metrics':'list',
                            'plot_post_pred':'bool',
                            'plot_map_pred':'bool',
                            'max_a_posteriori':'string',
                            'corner_title_fmt':'string',
                            'out_dir':'string',
                            'bmc':'bool'
                            },
                        methods = [
                            'setup_inference()','create_inference()','create_error_model()',
                            'perform_bootstrap()', '_perturb_data()', '_eval_model()',
                            'normpdf()', '_coor_Factor_BME()', '_rejection_sampling()',
                            '_posterior_predictive()', '_plot_max_a_posteriori()',
                            'plot_post_params()', 'plot_log_BME()', '_plot_post_predictive()'
                            ]
                        )
    graph.add_class(bayesinf)
    graph.add_composition(engine, bayesinf, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(disc, bayesinf, multiplicity_parent = 1, multiplicity_child = 1)
    
    # MCMC class 
    mcmc = UMLClass('MCMC', 
                    attributes = {
                        'engine':'Engine', 
                        'mcmc_params':'dict', 
                        'Discrepancy':'Discrepancy', 
                        'bias_inputs':'dict', 
                        'error_model':'MetaModel', 
                        'req_outputs':'list', 
                        'selected_indices':'array', 
                        'emulator':'bool',
                        'out_dir':'string', 
                        'name':'string',
                        },
                    methods = [
                        'run_sampler()', 'log_prior()', 'log_likelihood()',
                        'log_posterior()', 'eval_model()', 'train_error_model()',
                        'normpdf()'
                        ]
                    )
    graph.add_class(mcmc)
    graph.add_composition(mcmc, bayesinf, multiplicity_parent=1, multiplicity_child=1)
    graph.add_composition(disc, mcmc, multiplicity_parent=1, multiplicity_child=1)
    
    print_and_save(graph, 'bayesian_validation')
                        

def generate_bmc_uml():
    graph = Graph('pyUML')
    
    expdesign = UMLClass('ExpDesigns')
    graph.add_class(expdesign)
    
    model = UMLClass('PyLinkForwardModel')
    graph.add_class(model)
    
    metamod = UMLClass('MetaModel')
    graph.add_class(metamod)
    
    engine = UMLClass('Engine')
    graph.add_class(engine)
    graph.add_composition(model, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(expdesign, engine, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(metamod, engine, multiplicity_parent = 1, multiplicity_child = 1)
    
    disc = UMLClass('Discrepancy')
    graph.add_class(disc)
    
    bayesinf = UMLClass('BayesInference')
    graph.add_class(bayesinf)
    graph.add_composition(engine, bayesinf, multiplicity_parent = 1, multiplicity_child = 1)
    graph.add_composition(disc, bayesinf, multiplicity_parent = 1, multiplicity_child = 1)
    
    # MCMC class 
    mcmc = UMLClass('MCMC')
    graph.add_class(mcmc)
    graph.add_composition(mcmc, bayesinf, multiplicity_parent=1, multiplicity_child=1)
    graph.add_composition(disc, mcmc, multiplicity_parent=1, multiplicity_child=1)
    
    
    # BMC
    bmc = UMLClass('BayesianModelComparison',
                   attributes = {'justifiability':'bool', 
                                 'perturbed_data':'array',
                                 'n_bootstrap':'int', 
                                 'data_noise_level':'float',
                                 'use_Bayes_settings':'bool', 
                                 'emulator':'bool', 
                                 'out_dir':'string'
                                 },
                   methods = ['setup()', 'model_comparison_all()','calc_bayes_factors',
                              'calc_model_weights()', 'calc_justifiability_analysis()',
                              'generate_dataset()', 'perturb_data()','cal_model_weight()',
                              'plot_just_analysis()','plot_model_weights()', 'plot_bayes_factor()']
                   )
    graph.add_class(bmc)
    graph.add_composition(bayesinf, bmc, multiplicity_parent=1, multiplicity_child='1..*')
    
    print_and_save(graph, 'bayesian_model_comparison')
                        

if __name__ == '__main__':
    
    generate_input_uml()
    generate_input_uml_reduced()
    generate_model_uml()
    generate_training_uml()
    generate_training_uml_reduced()
    generate_al_uml()
    generate_al_uml_reduced()
    generate_postprocessing_uml()
    generate_bayes_uml()
    generate_bmc_uml()